from datetime import datetime, timedelta
import os
import re
import subprocess
import configparser

from pathlib import Path

from pydantic import BaseModel, field_validator, model_validator
from typing import Optional, Union, List, Tuple, ClassVar


class BaseAttribute(BaseModel):
    """Abstract base class for attribute configurations."""
    val: str

    command: ClassVar[str] = ""

    available: ClassVar[List[str]] = []
    current: ClassVar[Optional[str]] = None

    def apply(self):
        if self.val and self.val != self.current:
            subprocess.run([self.command, self.val])

    @classmethod
    def fetch_themes(cls, cmd: str,
                     regex: str) -> Tuple[List[str], Optional[str]]:
        """Fetch available and the current theme."""
        if cls.available and cls.current:
            return cls.available, cls.current

        try:
            output = subprocess.run(cmd.split(),
                                    capture_output=True,
                                    text=True,
                                    check=True).stdout.strip()

            for line in output.splitlines():
                match = re.search(regex, line)
                if match:
                    cls.available.append(match.group(1))
                    if "current" in line.lower():
                        cls.current = match.group(1)

            return cls.available, cls.current
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Failed to fetch themes: {e}")

    @classmethod
    def init_themes(cls, fetch_function):
        """Generic initialization for themes."""
        cls.available, cls.current = fetch_function()

    def validate_theme(self):
        if self.val and self.val not in self.available:
            raise ValueError(
                f"Invalid attribute: {self.val}. Available options are {self.available}."
            )

        return self

    def __str__(self) -> str:
        return self.val


class Colorscheme(BaseAttribute):
    available: ClassVar[List[str]] = []
    current: ClassVar[Optional[str]] = None

    command = "plasma-apply-colorscheme"

    @classmethod
    def fetch_colorschemes(cls) -> Tuple[List[str], Optional[str]]:
        """Fetch available colorschemes and the current colorscheme."""
        return cls.fetch_themes(f"{cls.command} -l", r" \* ([\w\s\-]+\w)")

    @model_validator(mode="after")
    def init_colorschemes(self):
        """Initialization of colorscheme variables."""

        self.init_themes(self.fetch_colorschemes)
        return self

    @model_validator(mode="after")
    def validate_colorscheme(self):
        self.validate_theme()
        return self


class CursorTheme(BaseAttribute):
    available: ClassVar[List[str]] = []
    current: ClassVar[Optional[str]] = None

    command = "plasma-apply-cursortheme"

    @classmethod
    def fetch_cursorthemes(cls) -> Tuple[List[str], Optional[str]]:
        """Fetch available cursorthemes and the current cursortheme."""
        return cls.fetch_themes(f"{cls.command} --list-themes",
                                r"\* .* \[(.*?)\]")

    @model_validator(mode="after")
    def init_cursorthemes(self):
        self.init_themes(self.fetch_cursorthemes)
        return self

    @model_validator(mode="after")
    def validate_colorscheme(self):
        self.validate_theme()
        return self


class DesktopTheme(BaseAttribute):
    available: ClassVar[List[str]] = []
    current: ClassVar[Optional[str]] = None

    command = "plasma-apply-desktoptheme"

    @classmethod
    def fetch_desktopthemes(cls) -> Tuple[List[str], Optional[str]]:
        """Fetch available desktopthemes and the current desktoptheme."""
        return cls.fetch_themes(f"{cls.command} --list-themes",
                                r" \* ([\w-]+)")

    @model_validator(mode="after")
    def init_desktopthemes(self):
        self.init_themes(self.fetch_desktopthemes)
        return self

    @model_validator(mode="after")
    def validate_colorscheme(self):
        self.validate_theme()
        return self


class IconTheme(BaseAttribute):
    available: ClassVar[List[str]] = []
    current: ClassVar[Optional[str]] = None

    @classmethod
    def fetch_iconthemes(cls) -> Tuple[List[str], Optional[str]]:
        if cls.available and cls.current:
            return cls.available, cls.current

        home_dir = Path.home()
        old_icon_dir = home_dir / ".icons"
        icon_dir = home_dir / ".local/share/icons"
        system_icon_dir = Path("/usr/share/icons")

        for path in [old_icon_dir, icon_dir, system_icon_dir]:
            if path.exists():
                cls.available += [
                    item.name for item in path.iterdir() if item.is_dir()
                ]

        kdeconfig_path = Path.home() / ".config/kdeglobals"

        if kdeconfig_path.exists():
            config = configparser.ConfigParser()
            config.read(kdeconfig_path)

            if config.has_section("Icons"):
                cls.current = config.get("Icons", "Theme")
            else:
                cls.current = None

        return cls.available, cls.current

    @model_validator(mode="after")
    def init_iconthemes(self):
        if not self.command:

            search_paths = [
                "/usr/local/libexec", "/usr/local/lib", "/usr/libexec",
                "/usr/lib", "/usr/lib/x86_64-linux-gnu/libexec",
                "/usr/lib/aarch64-linux-gnu/libexec"
            ]

            for path in search_paths:
                executable_path = Path(path) / "plasma-changeicons"
                if executable_path.is_file():
                    IconTheme.command = str(executable_path)

        self.fetch_iconthemes()
        return self

    @model_validator(mode="after")
    def validate_colorscheme(self):
        self.validate_theme()
        return self


class Wallpaper(BaseAttribute):
    path: Optional[Path] = None
    command = "plasma-apply-wallpaperimage"
    available: ClassVar[List[str]] = []
    current: ClassVar[Optional[str]] = None

    @classmethod
    def fetch_wallpapers(cls) -> Tuple[List[str], Optional[str]]:
        if cls.available and cls.current:
            return cls.available, cls.current

        config_file = Path(
            '~/.config/plasma-org.kde.plasma.desktop-appletsrc').expanduser()
        cls.current = ""

        # Open and search for the Image variable
        with open(config_file, 'r') as file:
            section_found = False
            for line in file:
                # Look for the specific section
                if '[Wallpaper][org.kde.image]' in line:
                    section_found = True

                # Look for the Image variable after finding the section
                elif section_found and line.strip().startswith('Image='):
                    cls.current = line.strip().split('=', 1)[1]
                    cls.current = cls.current.replace('file://', '')
                    break

        cls.available = []
        valid_extensions = {
            '.jpg', '.jpeg', '.jxl', '.png', '.bmp', '.webp', '.tiff'
        }

        for d in ["~/.local/share/wallpapers/", "/usr/share/wallpapers"]:
            path = Path(d).expanduser()
            if not path.exists():
                continue

            for entry in path.iterdir():

                if entry.is_dir():  # Check for metadata.json in directories
                    metadata_path = entry / "metadata.json"
                    if metadata_path.is_file():
                        cls.available.append(str(entry))

                elif entry.is_file():  # Check for valid extensions in files
                    if entry.suffix.lower() in valid_extensions:
                        cls.available.append(str(entry))

        return cls.available, cls.current

    @model_validator(mode="after")
    def init_wallpaper(self):
        self.init_themes(self.fetch_wallpapers)

        name_to_path = {Path(p).name: Path(p) for p in self.available}

        if self.val:
            if self.val in name_to_path:
                self.path = name_to_path[self.val]
            else:
                self.path = Path(self.val).expanduser()

        if self.path and self.path.exists():
            self.val = str(self.path)

            if self.val not in self.available:
                self.available.append(self.val)

        return self

    @model_validator(mode="after")
    def validate_colorscheme(self):
        self.validate_theme()
        return self


class Theme(BaseModel):
    colorscheme: Optional[Colorscheme] = None
    cursortheme: Optional[CursorTheme] = None
    desktoptheme: Optional[DesktopTheme] = None
    icontheme: Optional[IconTheme] = None
    wallpaper: Optional[Wallpaper] = None

    command: Optional[str] = None
    time: List[Union[str, datetime]] = []
    enabled: bool = True

    def __str__(self) -> str:
        components = {}

        for attr in [
                "colorscheme", "cursortheme", "desktoptheme", "icontheme",
                "wallpaper"
        ]:
            if eval(f"self.{attr}"):
                components[attr] = eval(f"self.{attr}.val")

        if self.time:
            time_str = [
                t.strftime("%H:%M") if isinstance(t, datetime) else t
                for t in self.time
            ]
            components["time"] = time_str

        components["enabled"] = self.enabled

        if self.command:
            components["command"] = self.command

        # Format the output like YAML
        result = "\n    ".join(f"{key}: {value}"
                               for key, value in components.items())
        return result

    def kshift(self) -> None:

        for attr in [
                self.colorscheme, self.cursortheme, self.desktoptheme,
                self.icontheme, self.wallpaper
        ]:
            if attr:
                attr.apply()

        if self.command:
            os.system(self.command)

    @model_validator(mode="before")
    def parse_attributes(cls, values):
        mtch = {
            "colorscheme": Colorscheme,
            "cursortheme": CursorTheme,
            "desktoptheme": DesktopTheme,
            "icontheme": IconTheme,
            "wallpaper": Wallpaper
        }

        for attr in mtch.keys():
            if attr in values and isinstance(values[attr], str):
                attr_cls = mtch[attr]
                values[attr] = attr_cls(val=values[attr])

        return values

    @field_validator("time", mode="before")
    def parse_time(cls, value):
        if isinstance(value, str):
            return [value]  # Convert single string to a list
        elif isinstance(value, list):
            if not all(isinstance(item, str) for item in value):
                raise ValueError(
                    "All elements in the 'time' list must be strings.")
            return value
        else:
            raise TypeError("'time' must be a string or a list of strings.")

    @model_validator(mode="after")
    def if_disabled(self):
        if not self.enabled:
            self.time = []

        return self

    @field_validator("time", mode="after")
    def convert_time_strings(cls, times):
        """HH:MM times are converted to datetime, other strings are to be checked if OnCalendar"""
        new_times = []
        for item in times:
            if isinstance(item, str):
                if re.match(r'^\d{2}:\d{2}$', item):
                    # Parse "HH:MM" into a datetime object with today's date
                    now = datetime.now()
                    hour, minute = map(int, item.split(':'))
                    dt = now.replace(hour=hour,
                                     minute=minute,
                                     second=0,
                                     microsecond=0)
                    # If the time has already passed today, schedule for tomorrow
                    if dt < now:
                        dt += timedelta(days=1)
                    item = dt

                new_times.append(item)
            elif isinstance(item, datetime):
                new_times.append(item)
            else:
                raise TypeError(
                    "Each time entry must be a string in HH:MM or systemd OnCalendar format."
                )

        return new_times



from datetime import datetime


def test_get_sundata_real_api(mocker):
    from kshift.theme import Colorscheme

    mocker.patch.object(Colorscheme, 'available',
                        ["BreezeClassic", "BreezeDark", "BreezeLight"])
    mocker.patch.object(Colorscheme, 'current', "BreezeDark")

    from kshift.conf import Config, defaults

    config = Config()

    default_sunrise = datetime.strptime(defaults["sunrise"], "%H:%M")
    default_sunset = datetime.strptime(defaults["sunset"], "%H:%M")

    # Test fetching sunrise data from the real API
    sunrise = config.get_sundata("sunrise")
    sunset = config.get_sundata("sunset")

    # Assert that the returned sunrise and sunset times are valid datetime objects
    assert sunrise.hour >= 0 and sunrise.hour <= 23
    assert sunset.hour >= 0 and sunset.hour <= 23
    assert sunrise < sunset  # Sunrise must occur before sunset

    assert sunrise.time != default_sunrise.time
    assert sunset.time != default_sunset.time



from datetime import datetime, timedelta
import os
import yaml
import re
import colorama
import subprocess
import requests
import json

from kshift.theme import Theme

from pathlib import Path

from pydantic import BaseModel, Field, field_validator, model_validator
from typing import Dict

defaults = {
    "latitude": 39,
    "longitude": -77,
    "sunrise": "08:00",
    "sunset": "18:00",
    "rise_delay": 0,
    "set_delay": 0,
    "net_timeout": 10,
    "themes": {
        'day': Theme(colorscheme="BreezeLight", time=["sunrise"]),
        'night': Theme(colorscheme="BreezeDark", time=["sunset"]),
    }
}


class Config(BaseModel):
    # YAML Variables
    latitude: float = Field(
        defaults["latitude"],
        ge=-90,
        le=90,
        description="Latitude coordinate for sun position data")
    longitude: float = Field(
        defaults["longitude"],
        ge=-180,
        le=180,
        description="Longitude Coordinate for sun position data")

    sunrise: datetime = Field(
        datetime.strptime(defaults["sunrise"], "%H:%M"),
        description="Default sunrise time in HH:MM format.")
    sunset: datetime = Field(
        datetime.strptime(defaults["sunset"], "%H:%M"),
        description="Default sunset time in HH:MM format.")

    rise_delay: int = Field(
        defaults["rise_delay"],
        ge=-23,
        le=23,
        description="Hour delay for sunrise, between -23 and 23.")
    set_delay: int = Field(
        defaults["set_delay"],
        ge=-23,
        le=23,
        description="Hour delay for sunset, between -23 and 23.")

    webdata: bool = Field(
        True, description="Whether to fetch sunrise/sunset data from the web.")
    net_timeout: int = Field(
        defaults["net_timeout"],
        ge=0,
        le=60,
        description="Network timeout in seconds, between 0 and 60.")
    themes: Dict[str, Theme] = Field(defaults["themes"],
                                     description="Dictionary of themes.")

    # Environment-based paths
    home_directory: Path = Field(default=Path.home(),
                                 description="The user's home directory.")
    xdg_config: Path = Field(
        default=Path(os.getenv("XDG_CONFIG_HOME",
                               Path.home() / ".config")),
        description="Base directory for configuration files.",
    )
    xdg_data: Path = Field(
        default=Path(os.getenv("XDG_DATA_HOME",
                               Path.home() / ".local/share")),
        description="Base directory for application data.",
    )
    xdg_cache: Path = Field(
        default=Path(os.getenv("XDG_CACHE_HOME",
                               Path.home() / ".cache")),
        description="Base directory for cache data.",
    )

    # Dependent fields
    systemd_loc: Path = Field(
        Path.home(),
        description="Directory for systemd user timers and services.")
    config_loc_base: Path = Field(
        Path.home(), description="Base configuration directory for kshift.")
    config_loc: Path = Field(
        Path.home(), description="Path to the main kshift configuration file.")
    log_loc: Path = Field(Path.home(), description="Path to kshift log file.")

    sun_api: str = Field("",
                         description="URL to fetch sunrise and sunset data.")
    api_file: Path = Field(
        Path.home(),
        description="Path to the cache file for sunrise and sunset data.")

    @model_validator(mode="after")
    def set_dependant_paths(self):
        # Compute dependent paths
        self.systemd_loc = self.xdg_data / "systemd/user"
        self.config_loc_base = self.xdg_config / "kshift"
        self.config_loc = self.config_loc_base / "kshift.yml"
        self.sun_api = f"https://api.sunrisesunset.io/json?lat={self.latitude}&lng={self.longitude}"
        self.api_file = self.xdg_cache / "kshift" / f"{self.latitude}{self.longitude}.out"

        return self

    @field_validator("sunrise", "sunset", mode="before")
    def validate_time_format(cls, value) -> datetime:
        try:
            value = datetime.strptime(value, "%H:%M")
            value = datetime.combine(datetime.now().date(), value.time())
        except ValueError:
            raise ValueError(
                f"Invalid time format for '{value}'. Use 'HH:MM'.")

        return value

    @model_validator(mode='after')
    def parse_sun_times(self):

        def apply_delay(time_obj: datetime, delay_hours: int) -> datetime:
            return time_obj + timedelta(hours=delay_hours)

        for name, config in self.themes.items():

            updated_times = []
            for t in config.time:
                if isinstance(t, str):
                    if t == "sunrise":
                        t = apply_delay(self.get_sundata(t), self.rise_delay)
                    elif t == "sunset":
                        t = apply_delay(self.get_sundata(t), self.set_delay)
                    else:

                        # Determine if theme time is a valid calendar time
                        if t:
                            process = subprocess.run(
                                ["systemd-analyze", "calendar", t],
                                stdout=subprocess.PIPE)
                            stat_code = process.returncode

                            if stat_code == 0:
                                calendar_time = ""
                                output = process.stdout.decode('utf-8').strip()

                                for line in output.splitlines():
                                    r = re.search("Normalized form: (.*)",
                                                  line)
                                    if r:
                                        calendar_time = r.group(1)
                                        break

                                t = calendar_time

                            else:
                                raise ValueError(
                                    "Invalid systemd calendar time!: " + t)

                    updated_times.append(t)
                elif isinstance(t, datetime):
                    updated_times.append(t)
                else:
                    raise ValueError(
                        f"Unsupported time in theme '{name}': {t}")

            # Update the theme's time with resolved datetime objects
            config.time = updated_times

        return self

    # Prints the status of Kshift, the timer, and the current config file
    def status(self):
        if self.systemd_loc.exists() and self.config_loc.exists():

            timer_status_cmd = "systemctl --user is-enabled "
            enabled = False
            timed_outputs = ""

            for f in os.listdir(self.systemd_loc):
                if 'kshift-' in f and '.timer' in f:
                    timer_status = subprocess.run(
                        (timer_status_cmd + f).split(),
                        stdout=subprocess.PIPE).stdout.decode('utf-8').strip()

                    if timer_status == "enabled":
                        enabled = True

                        theme_name = f.replace(".timer", "").split("-")[1]

                        timer = open(self.systemd_loc / f, "r")
                        for line in timer:
                            m = re.search("OnCalendar=(.*)", line)
                            if m:
                                time = m.group(1)
                                timed_outputs += f"{theme_name:<10} {time}\n"

                        timer.close()

            if enabled:
                print("kshift status: " + colorama.Fore.GREEN + "ENABLED" +
                      colorama.Fore.WHITE)
                print(timed_outputs, end='')
            else:
                print("kshift status: " + colorama.Fore.RED + "DISABLED." +
                      colorama.Fore.WHITE)
        else:
            print(
                "kshift not installed. Run again with `--install` and edit your config with `--config`"
            )

    def _select_sunstate(self, sunstate) -> datetime:
        if sunstate == "sunrise":
            return self.sunrise
        elif sunstate == "sunset":
            return self.sunset
        else:
            raise ValueError(
                f"Invalid sunstate '{sunstate}'. Use 'sunrise' or 'sunset'.")

    # Gets the sundata from the internet and writes it to a tmp file
    #
    # Sets sunrise and sunset
    # Returns the correct sunstate
    def web_sundata(self, sunstate):
        url = self.sun_api
        try:
            response = requests.get(url, timeout=self.net_timeout)
            response.raise_for_status()
            data = response.json()

            sunrise = datetime.strptime(data["results"]["sunrise"],
                                        "%I:%M:%S %p").time()
            sunset = datetime.strptime(data["results"]["sunset"],
                                       "%I:%M:%S %p").time()

            self.sunrise = datetime.combine(datetime.now().date(), sunrise)
            self.sunset = datetime.combine(datetime.now().date(), sunset)

            cache_data = {
                "location": f"{self.latitude},{self.longitude}",
                "sunrise": self.sunrise.isoformat(),
                "sunset": self.sunset.isoformat()
            }

            cache_dir = self.api_file.parent
            cache_dir.mkdir(parents=True, exist_ok=True)
            with open(self.api_file, "w") as file:
                json.dump(cache_data, file)

        except requests.exceptions.ConnectionError as e:
            print(
                f"Connection error: {e}. Could not connect to {url}. Falling back to defaults."
            )
        except requests.exceptions.Timeout as e:
            print(
                f"Timeout error: {e}. The request to {url} took too long. Falling back to defaults."
            )
        except requests.exceptions.HTTPError as e:
            print(
                f"HTTP error: {e}. Invalid response from the server. Falling back to defaults."
            )
        except json.JSONDecodeError as e:
            print(
                f"JSON decoding error: {e}. Invalid response format. Falling back to defaults."
            )
        except Exception as e:
            print(f"Unexpected error: {e}. Falling back to defaults.")

        return self._select_sunstate(sunstate)

    # Checks to see if sundata is in the designated tmp file, if not, it calls web_sundata
    # Returns the correct sunstate
    def get_sundata(self, sunstate):
        if self.webdata is False:
            return self._select_sunstate(sunstate)

        elif self.api_file.exists():
            try:
                with open(self.api_file, "r") as file:
                    cache_data = json.load(file)
                    cache_data["sunrise"] = datetime.fromisoformat(
                        cache_data["sunrise"])
                    cache_data["sunset"] = datetime.fromisoformat(
                        cache_data["sunset"])

                if cache_data[
                        "location"] == f"{self.latitude},{self.longitude}":

                    self.sunrise = datetime.combine(
                        datetime.now().date(), cache_data["sunrise"].time())
                    self.sunset = datetime.combine(datetime.now().date(),
                                                   cache_data["sunset"].time())

                if cache_data["sunrise"].date() == datetime.today().date(
                ) and cache_data["sunset"].date() == datetime.today().date():
                    return self._select_sunstate(sunstate)

            except (ValueError, IndexError):
                pass  # Fall back to fetching data if the cache is corrupted

        # Fetch fresh data if the cache is invalid or missing
        return self.web_sundata(sunstate)


def load_config() -> Config:
    # Determine the config file path
    config_base = Path.home() / ".config/kshift"
    config_file = config_base / "kshift.yml"

    config_data = defaults

    # If user configuration exists, overwrite the defaults
    if config_file.exists():
        with open(config_file, "r") as file:
            try:
                user_data = yaml.safe_load(file)
                config_data.update(user_data)  # Merge user data into defaults
            except yaml.YAMLError as e:
                print(f"Error reading kshift.yml: {e}")
                raise
    else:
        print(
            f"User configuration file not found at {config_file}. Using defaults."
        )

    # Instantiate and return the Config object
    return Config(**config_data)



#!/usr/bin/env python

import click

from datetime import datetime
from os import system, makedirs, getenv
from re import search
import subprocess
from shutil import which

import logging
from logging.handlers import RotatingFileHandler
import json

from importlib.resources import files

from kshift.conf import load_config

from string import Template

from kshift.theme import *

c = load_config()

###################################
# Logging
###################################

log_file = c.xdg_cache / "kshift" / "kshift.log"
cache_dir = log_file.parent
cache_dir.mkdir(parents=True, exist_ok=True)

# Setup RotatingFileHandler for the same file
handler = RotatingFileHandler(log_file,
                              maxBytes=1 * 1024 * 1024,
                              backupCount=1)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s",
                              datefmt="%Y-%m-%d %H:%M:%S")
handler.setFormatter(formatter)

# Configure logging
logging.basicConfig(level=logging.INFO, handlers=[handler])


def log_theme_change(theme_name):
    """
    Log a theme change event with structured data.
    """
    log_data = {
        "event": "theme_change",
        "theme": theme_name,
        "source": getenv("SOURCE", "direct")
    }
    logging.info(json.dumps(log_data))  # Use JSON for structured logs


def log_element_change(theme: Theme):
    log_data = {"event": "specific_change", "theme": str(theme)}
    logging.info(json.dumps(log_data))


def log_timer_write(themes):
    log_data = {"event": "timers_written", "themes": themes}
    logging.info(json.dumps(log_data))


def parse_theme_logs(log_file, reference_time=None):
    """
    Parse the log file to determine the last activated theme.
    """

    last_theme = None
    reference_time = reference_time or datetime.now()

    with open(log_file, "r") as f:
        # for line in f:
        for line in reversed(f.readlines()):
            try:
                # Extract JSON from log message
                timestamp, _, json_data = line.partition("- INFO - ")
                log_entry = json.loads(json_data)

                # Check if this log entry is a theme change event
                if log_entry.get("event") == "theme_change" and log_entry.get(
                        "source") == "systemd" and c.themes.get(
                            log_entry.get("theme")):

                    log_time = datetime.fromisoformat(timestamp)
                    return (log_entry["theme"], log_time)

            except (json.JSONDecodeError, ValueError):
                # Skip lines that aren't properly formatted
                continue

    return last_theme


###################################
# systemd
###################################


# Writes the timers/services for each timed theme
def write_systemd():

    def write_timer(path, subs):
        # write theme timer
        with open(path, "w") as file:
            template = Template(
                open(c.config_loc_base / "templates/template.timer").read())

            file.write(template.substitute(subs))

    def write_service(path, subs):
        # write theme timer
        with open(path, "w") as file:
            template = Template(
                open(c.config_loc_base / "templates/template.service").read())

            file.write(template.substitute(subs))

    kshift_path = which("kshift") or "kshift"
    written_timers = []

    # Remove any old timers
    for name, conf in c.themes.items():
        timer = c.systemd_loc / f"kshift-{name}.timer"
        if not conf.time and timer.exists():
            timer.unlink()

    # Write services for each theme
    # Write timer if they have 'time' option
    for (name, conf) in c.themes.items():

        theme_service_str = " ".join([
            f'kshift-{x[0]}.service' for x in c.themes.items()
            if x[0] is not name
        ])

        # Convert any datetime objects to simple HH:MM
        # Datetime objects were originally in HH:MM or sunset/sunrise
        # If it is a string, it was a verified OnCalendar time
        theme_times = []
        for t in conf.time:
            if isinstance(t, datetime):
                t = datetime.strftime(t, "%H:%M")

            theme_times.append(t)

        # Extract old timer times to check if the timer needs to be updated
        timer_path = (c.systemd_loc / f"kshift-{name}.timer")
        timer_times = []

        timer_changed = False
        if timer_path.exists():

            for line in open(timer_path, "r"):
                r = search("OnCalendar=(.*)\n", line)
                if r:
                    timer_times.append(r.group(1))

            timer_changed = timer_times != theme_times

        if not timer_path.exists() or timer_changed:

            calendar_times = ""
            for t in theme_times:
                calendar_times += f'OnCalendar={t}\n'

            # Cannot have timer with no timer action
            if calendar_times:
                subs = {
                    "description": f'kshift timer for theme {name}',
                    "unit_options": f'After={theme_service_str}',
                    "timer_action": f'{calendar_times}\nPersistent=true'
                }
                write_timer(timer_path, subs)
                written_timers.append(name)

            subs = {
                "description": f'kshift service for theme {name}',
                "command": f"{kshift_path} theme {name}"
            }
            write_service(c.systemd_loc / f"kshift-{name}.service", subs)

    # write startup timer & service
    startup_timer = c.systemd_loc / "kshift-startup.timer"
    startup_service = c.systemd_loc / "kshift-startup.service"
    if not startup_timer.exists() or not startup_service.exists():

        subs = {
            "description": 'kshift startup timer',
            "unit_options": "",
            "timer_action": 'OnStartupSec=5'
        }
        write_timer(startup_timer, subs)
        written_timers.append("startup")

        subs = {
            "description": 'kshift startup service',
            "command": f"{kshift_path}"
        }
        write_service(startup_service, subs)

    if len(written_timers) > 0:
        log_timer_write(written_timers)

        timers_str = ""
        for t in written_timers:
            timers_str += f"kshift-{t}.timer "

        system("systemctl --user daemon-reload")
        system(f"systemctl --user enable { timers_str }")
        system(f"systemctl --user start { timers_str }")


###################################
# CLI
###################################


@click.group(
    invoke_without_command=True,
    help="KDE Theme Switching (kshift)",
)
@click.pass_context
def cli(ctx):
    """Main entry point for the kshift CLI."""
    if ctx.invoked_subcommand is None:
        # Call the theme subcommand if no subcommand is provided
        ctx.invoke(theme)


# Installs systemd services and timers
@cli.command(help="Install kshift systemd services")
def install():

    answer = input("Are you sure you want to install kshift? [y/n]: ")
    if answer in ("Y", "y", "yes"):

        if not c.config_loc_base.exists():
            c.config_loc_base.mkdir()

        if not (c.config_loc_base / "templates").exists():
            system(
                f'cp -r {files("kshift") / "templates"} {c.config_loc_base}')

        if not c.config_loc.exists():
            system(
                f"cp {files('kshift') / 'defaults.yml'} {c.config_loc_base / 'kshift.yml'}"
            )

        makedirs(c.systemd_loc, exist_ok=True)

        print("Writing & enabling theme timers and services...")
        write_systemd()


# Removes kshift timer
@cli.command(help="Remove kshift systemd services")
def remove():

    answer = input("Are you sure you want to remove kshift? [y/n]: ")
    if answer in ("Y", "y", "yes"):
        print("Removing kshift timers and services...")

        system(f"rm {c.systemd_loc / 'kshift-'}*")

        system("systemctl --user daemon-reload")


@cli.command(help="Display kshift status")
def status():
    c.status()


@cli.command(help="Edit the kshift configuration file")
def config():
    """Edit the configuration file."""

    filepath = c.config_loc
    if filepath.exists():
        try:
            # Use xdg-open to open the file in the default editor
            subprocess.run(["xdg-open", filepath], check=True)
            print(f"Opened {filepath} in the default editor.")
        except subprocess.CalledProcessError as e:
            print(f"Failed to open the file: {e}")
    else:
        print(f"Config does not exist @ {filepath}")


@cli.command(help="Tail on kshift logs")
@click.option(
    "-a",
    "--all",
    is_flag=True,
    help="Print the entire log file",
)
def logs(all):

    print(f"Log @ {log_file}")
    with open(log_file, 'r') as f:
        lines = f.readlines()
        if not all:
            lines = lines[-10:]

        for line in lines:
            print(line, end="")


@cli.command(help="List possible themes or attributes")
@click.argument("attribute",
                type=click.Choice([
                    "themes", "colorschemes", "cursorthemes", "desktopthemes",
                    "iconthemes", "wallpapers"
                ],
                                  case_sensitive=False))
def list(attribute):

    def print_available(attr, items):
        print(f"Available {attr}:")
        for a in items:
            print(f"- {a}")

    if attribute == "themes":
        for name, conf in c.themes.items():
            print(f"theme: {name}\n    {conf}\n")
            pass

    items = []
    match attribute:
    # items are the class.available
        case "colorschemes":
            items = Colorscheme.fetch_colorschemes()[0]
        case "cursorthemes":
            items = CursorTheme.fetch_cursorthemes()[0]
        case "desktopthemes":
            items = DesktopTheme.fetch_desktopthemes()[0]
        case "iconthemes":
            items = IconTheme.fetch_iconthemes()[0]
        case "wallpapers":
            items = Wallpaper.fetch_wallpapers()[0]

    if items:
        print_available(attribute, items)


@cli.command(help="Change themes or apply specific theme elements")
@click.argument("theme",
                type=str,
                required=False,
                default=None,
                nargs=1,
                metavar="[THEME_NAME]")
@click.option(
    "-csr",
    "--cursortheme",
    type=str,
    help="Set a specific cursor theme (overrides theme)",
)
@click.option(
    "-c",
    "--colorscheme",
    type=str,
    help="Set a specific colorscheme (overrides theme)",
)
@click.option(
    "-dk",
    "--desktop_theme",
    type=str,
    help="Set a specific desktop theme (overrides theme)",
)
@click.option(
    "-i",
    "--icontheme",
    type=str,
    help="Set a specific icon theme (overrides theme)",
)
@click.option(
    "-w",
    "--wallpaper",
    type=str,
    help="Set a specific wallpaper (overrides theme)",
)
@click.option(
    "-dk",
    "--desktop_theme",
    type=str,
    help="Set a specific desktop theme (overrides theme)",
)
def theme(theme, colorscheme, cursortheme, desktop_theme, icontheme,
          wallpaper):

    kshift_status = subprocess.run(
        "systemctl --user is-enabled kshift-startup.timer".split(),
        stdout=subprocess.PIPE).stdout.decode('utf-8').strip()

    if theme:
        if theme in c.themes:
            print(f"Applying theme {theme}...")
            c.themes[theme].kshift()
            log_theme_change(theme)
        else:
            print(f"Error: Theme '{theme}' not found in configuration.")

    if any([colorscheme, cursortheme, desktop_theme, icontheme, wallpaper]):
        # Apply individual theme elements dynamically
        custom_theme = Theme(
            colorscheme=colorscheme,
            cursortheme=cursortheme,
            icontheme=icontheme,
            wallpaper=wallpaper,
            desktoptheme=desktop_theme,
        )
        print("Applying custom theme elements...")
        custom_theme.kshift()
        log_element_change(custom_theme)

    # If there were no arguments
    # Determine which theme should be active, then shift to it
    if not any([theme, colorscheme, icontheme, wallpaper, desktop_theme]):

        themes = []

        # Add each theme with simple timer, ie not strict OnCalendar time
        for name, conf in c.themes.items():
            for t in conf.time:
                if isinstance(t, datetime) and datetime.now() >= t:
                    themes.append((name, t))

        # The last theme activated by timer could be correct active theme
        # Find this last time only if kshift is enabled in systemd
        if kshift_status == "enabled":
            last_log_theme = parse_theme_logs(log_file)
            if last_log_theme:
                themes.append(last_log_theme)

        # Sort themes such that the one closest to present is last
        # Then change to this theme
        themes = sorted(themes, key=lambda x: x[1])
        if themes:
            curr_theme = themes[-1][0]
            print(f"Applying theme {curr_theme}...")

            c.themes[curr_theme].kshift()
            log_theme_change(curr_theme)

    # Update systemd if installed
    if kshift_status == "enabled":
        write_systemd()


if __name__ == "__main__":
    cli()



from datetime import datetime
from pathlib import Path
import subprocess
import re
import os
import configparser


# Gets the names of all available colorschemes
def get_colorschemes():
    arr = []

    colorscheme_cmd = "plasma-apply-colorscheme -l"
    output = subprocess.run(colorscheme_cmd.split(),
                            capture_output=True,
                            text=True).stdout.strip()

    for line in output.splitlines():
        r = re.search(" \\* ([A-Za-z]*)", line)
        if r:
            arr.append(r.group(1))

    return arr


# Gets the current colorscheme
def curr_colorscheme():

    curr = ""

    colorscheme_cmd = "plasma-apply-colorscheme -l"
    output = subprocess.run(
        colorscheme_cmd.split(),
        stdout=subprocess.PIPE).stdout.decode('utf-8').strip()

    for line in output.splitlines():
        r = re.search(" \\* ([A-Za-z]*) \\(current color scheme\\)", line)
        if r:
            curr = r.group(1)
            break

    return curr


def get_iconthemes():

    arr = []

    home_dir = Path.home()
    old_icon_dir = home_dir / ".icons"
    icon_dir = home_dir / ".local/share/icons"
    system_icon_dir = Path("/usr/share/icons")

    for path in [old_icon_dir, icon_dir, system_icon_dir]:
        if path.exists():
            arr += [item.name for item in path.iterdir() if item.is_dir()]

    return arr


# Gets the current icon theme
# TODO FIX, currently broken
def curr_icontheme():

    curr = ""
    kdeconfig_path = f"{os.path.expanduser('~')}/.config/kdeglobals"

    if os.path.exists(kdeconfig_path):
        config = configparser.ConfigParser()
        config.read(kdeconfig_path)
        curr = config["Icons"]["Theme"]

    return curr


# Gets the path to plasma-changeicon
def find_plasma_changeicons():
    search_paths = [
        "/usr/local/libexec", "/usr/local/lib", "/usr/libexec", "/usr/lib",
        "/usr/lib/x86_64-linux-gnu/libexec",
        "/usr/lib/aarch64-linux-gnu/libexec"
    ]

    for path in search_paths:
        executable_path = os.path.join(path, "plasma-changeicons")
        if os.path.isfile(executable_path):
            return executable_path

    return None


def time_to_systemd(time):

    if time is None:
        return time

    process = subprocess.run(["systemd-analyze", "calendar", time],
                             stdout=subprocess.PIPE)
    stat_code = process.returncode

    if stat_code == 0:
        calendar_time = ""
        output = process.stdout.decode('utf-8').strip()

        for line in output.splitlines():
            r = re.search("Normalized form: (.*)", line)
            if r:
                calendar_time = r.group(1)
                break

        return calendar_time

    else:
        raise ValueError("Invalid systemd calendar time!: " + time)


# Converting to today's version of datetime for theme comparison
def systemd_to_datetime(time, today: datetime) -> datetime:

    if time is None or type(time).__name__ == "datetime":
        return time

    parts = time.split()

    # Y-M-D HH:MM:SS
    # OR
    # DOW Y-M-D HH:MM:SS

    if len(parts) == 3:
        time = " ".join(parts[1:])

    r = re.search("(.*)-(.*)-(.*) (.*):(.*):(.*)", time)
    if r is None:
        raise Exception("Bad systemd time format")

    today_broken = [
        today.year, today.month, today.day, today.hour, today.minute,
        today.second
    ]
    final = []

    for i in range(6):
        final.append(
            r.group(i + 1) if r.group(i + 1) != '*' else today_broken[i])

    if len(parts) == 3:
        dow = parts[0]
        new_time = "{} {}-{}-{} {}:{}:{}".format(dow, *final)

        date = datetime.strptime(new_time, "%a %Y-%m-%d %H:%M:%S")
    else:
        new_time = "{}-{}-{} {}:{}:{}".format(*final)

        date = datetime.strptime(new_time, "%Y-%m-%d %H:%M:%S")

    return date


# Gets the names of all available desktop themes
def get_desktopthemes():

    arr = []

    desktopthemes_cmd = "plasma-apply-desktoptheme --list-themes"
    output = subprocess.run(
        desktopthemes_cmd.split(),
        stdout=subprocess.PIPE).stdout.decode('utf-8').strip()

    for line in output.splitlines():
        r = re.search(" \\* ([A-Za-z]*(-[A-Za-z]*)?)", line)
        if r:
            arr.append(r.group(1))

    return arr


# Gets the current desktoptheme
def curr_desktoptheme():
    curr = ""

    desktoptheme_cmd = "plasma-apply-desktoptheme --list-themes"
    output = subprocess.run(
        desktoptheme_cmd.split(),
        stdout=subprocess.PIPE).stdout.decode('utf-8').strip()

    for line in output.splitlines():
        r = re.search(
            " \\* ([A-Za-z]*(-[A-Za-z]*)?) \\(current theme for the Plasma session\\)",
            line)
        if r:
            curr = r.group(1)
            break

    return curr


def open_in_default_editor(filepath: Path):
    if filepath.exists():
        try:
            # Use xdg-open to open the file in the default editor
            subprocess.run(["xdg-open", filepath], check=True)
            print(f"Opened {filepath} in the default editor.")
        except subprocess.CalledProcessError as e:
            print(f"Failed to open the file: {e}")
    else:
        print(f"File does not exist: {filepath}")
